@@ -1,20 +1,22 @@
 /*
-	libpe - the PE library
+    libpe - the PE library
 
-	Copyright (C) 2010 - 2013 libpe authors
+    Copyright (C) 2010 - 2015 libpe authors
+    
+    This file is part of libpe.
 
-	This program is free software: you can redistribute it and/or modify
-	it under the terms of the GNU General Public License as published by
-	the Free Software Foundation, either version 3 of the License, or
-	(at your option) any later version.
+    libpe is free software: you can redistribute it and/or modify
+    it under the terms of the GNU Lesser General Public License as published by
+    the Free Software Foundation, either version 3 of the License, or
+    (at your option) any later version.
 
-	This program is distributed in the hope that it will be useful,
-	but WITHOUT ANY WARRANTY; without even the implied warranty of
-	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-	GNU General Public License for more details.
+    libpe is distributed in the hope that it will be useful,
+    but WITHOUT ANY WARRANTY; without even the implied warranty of
+    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+    GNU Lesser General Public License for more details.
 
-	You should have received a copy of the GNU General Public License
-	along with this program.  If not, see <http://www.gnu.org/licenses/>.
+    You should have received a copy of the GNU Lesser General Public License
+    along with libpe.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "pe.h"
@@ -28,6 +30,11 @@
 #include <sys/mman.h>
 #include <sys/stat.h>
 
+bool pe_can_read(const pe_ctx_t *ctx, const void *ptr, size_t size) {
+	const uintptr_t end = (uintptr_t)LIBPE_PTR_ADD(ptr, size);
+	return ptr >= ctx->map_addr && end <= ctx->map_end;
+}
+
 pe_err_e pe_load_file(pe_ctx_t *ctx, const char *path) {
 	return pe_load_file_ext(ctx, path, 0);
 }
@@ -45,7 +52,8 @@ pe_err_e pe_load_file_ext(pe_ctx_t *ctx, const char *path, pe_options_e options)
 	}
 
 	// Open the file.
-	const int fd = open(ctx->path, O_RDWR);
+	int oflag = options & LIBPE_OPT_OPEN_RW ? O_RDWR : O_RDONLY;
+	const int fd = open(ctx->path, oflag);
 	if (fd == -1) {
 		//perror("open");
 		return LIBPE_E_OPEN_FAILED;
@@ -72,8 +80,11 @@ pe_err_e pe_load_file_ext(pe_ctx_t *ctx, const char *path, pe_options_e options)
 	ctx->map_size = stat.st_size;
 
 	// Create the virtual memory mapping.
-	ctx->map_addr = mmap(NULL, ctx->map_size, PROT_READ|PROT_WRITE,
-		MAP_SHARED, fd, 0);
+	int mprot = options & LIBPE_OPT_OPEN_RW ? PROT_READ|PROT_WRITE /* Pages may be written */ : PROT_READ;
+	// MAP_SHARED makes updates to the mapping visible to other processes that map this file.
+	// The file may not actually be updated until msync(2) or munmap() is called.
+	int mflags = options & LIBPE_OPT_OPEN_RW ? MAP_SHARED : MAP_PRIVATE;
+	ctx->map_addr = mmap(NULL, ctx->map_size, mprot, mflags, fd, 0);
 	if (ctx->map_addr == MAP_FAILED) {
 		close(fd);
 		//perror("mmap");
@@ -116,17 +127,12 @@ pe_err_e pe_unload(pe_ctx_t *ctx) {
 		fclose(ctx->stream);
 	}
 
-	if (ctx->path != NULL) {
-		free(ctx->path);
-	}
+	free(ctx->path);
+
 
 	// Dealloc internal pointers.
-	if (ctx->pe.directories != NULL) {
-		free(ctx->pe.directories);
-	}
-	if (ctx->pe.sections != NULL) {
-		free(ctx->pe.sections);
-	}
+	free(ctx->pe.directories);
+	free(ctx->pe.sections);
 
 	// Dealloc the virtual mapping.
 	if (ctx->map_addr != NULL) {
@@ -150,7 +156,7 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 
 	const uint32_t *signature_ptr = LIBPE_PTR_ADD(ctx->pe.dos_hdr,
 		ctx->pe.dos_hdr->e_lfanew);
-	if (LIBPE_IS_PAST_THE_END(ctx, signature_ptr, sizeof(uint32_t)))
+	if (!pe_can_read(ctx, signature_ptr, sizeof(uint32_t)))
 		return LIBPE_E_INVALID_LFANEW;
 
 	// NT signature (PE\0\0), or 16-bit Windows NE signature (NE\0\0).
@@ -167,8 +173,7 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 
 	ctx->pe.coff_hdr = LIBPE_PTR_ADD(signature_ptr,
 		LIBPE_SIZEOF_MEMBER(pe_file_t, signature));
-	if (LIBPE_IS_PAST_THE_END(ctx, ctx->pe.coff_hdr,
-		sizeof(IMAGE_COFF_HEADER)))
+	if (!pe_can_read(ctx, ctx->pe.coff_hdr, sizeof(IMAGE_COFF_HEADER)))
 		return LIBPE_E_MISSING_COFF_HEADER;
 
 	ctx->pe.num_sections = ctx->pe.coff_hdr->NumberOfSections;
@@ -179,7 +184,7 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 
 	// Figure out whether it's a PE32 or PE32+.
 	uint16_t *opt_type_ptr = ctx->pe.optional_hdr_ptr;
-	if (LIBPE_IS_PAST_THE_END(ctx, opt_type_ptr,
+	if (!pe_can_read(ctx, opt_type_ptr,
 		LIBPE_SIZEOF_MEMBER(IMAGE_OPTIONAL_HEADER, type)))
 		return LIBPE_E_MISSING_OPTIONAL_HEADER;
 
@@ -192,7 +197,7 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 			//fprintf(stderr, "ROM image is not supported\n");
 			return LIBPE_E_UNSUPPORTED_IMAGE;
 		case MAGIC_PE32:
-			if (LIBPE_IS_PAST_THE_END(ctx, ctx->pe.optional_hdr_ptr,
+			if (!pe_can_read(ctx, ctx->pe.optional_hdr_ptr,
 				sizeof(IMAGE_OPTIONAL_HEADER_32)))
 				return LIBPE_E_MISSING_OPTIONAL_HEADER;
 			ctx->pe.optional_hdr._32 = ctx->pe.optional_hdr_ptr;
@@ -203,7 +208,7 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 			ctx->pe.imagebase = ctx->pe.optional_hdr._32->ImageBase;
 			break;
 		case MAGIC_PE64:
-			if (LIBPE_IS_PAST_THE_END(ctx, ctx->pe.optional_hdr_ptr,
+			if (!pe_can_read(ctx, ctx->pe.optional_hdr_ptr,
 				sizeof(IMAGE_OPTIONAL_HEADER_64)))
 				return LIBPE_E_MISSING_OPTIONAL_HEADER;
 			ctx->pe.optional_hdr._64 = ctx->pe.optional_hdr_ptr;
@@ -228,8 +233,10 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 	ctx->pe.directories_ptr = LIBPE_PTR_ADD(ctx->pe.optional_hdr_ptr,
 		ctx->pe.optional_hdr.length);
 	// If there are no directories, sections_ptr will point right
-	// after the OPTIONAL header.
-	ctx->pe.sections_ptr = ctx->pe.directories_ptr;
+	// after the OPTIONAL header, otherwise, it will point right
+	// after the directories.
+	ctx->pe.sections_ptr = LIBPE_PTR_ADD(ctx->pe.directories_ptr,
+		ctx->pe.num_directories * sizeof(IMAGE_DATA_DIRECTORY));
 
 	if (ctx->pe.num_directories > 0) {
 		ctx->pe.directories = malloc(ctx->pe.num_directories
@@ -239,9 +246,6 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 		for (uint32_t i = 0; i < ctx->pe.num_directories; i++) {
 			ctx->pe.directories[i] = LIBPE_PTR_ADD(ctx->pe.directories_ptr,
 				i * sizeof(IMAGE_DATA_DIRECTORY));
-			// Calculate sections' start address.
-			ctx->pe.sections_ptr = LIBPE_PTR_ADD(ctx->pe.sections_ptr,
-				sizeof(IMAGE_DATA_DIRECTORY));
 		}
 	} else {
 		ctx->pe.directories_ptr = NULL;
@@ -263,6 +267,10 @@ pe_err_e pe_parse(pe_ctx_t *ctx) {
 	return LIBPE_E_OK;
 }
 
+bool pe_is_loaded(const pe_ctx_t *ctx) {
+	return ctx->map_addr != NULL && ctx->map_size > 0;
+}
+
 bool pe_is_pe(const pe_ctx_t *ctx) {
 	// Check MZ header
 	if (ctx->pe.dos_hdr == NULL || ctx->pe.dos_hdr->e_magic != MAGIC_MZ)
@@ -299,18 +307,38 @@ IMAGE_SECTION_HEADER *pe_rva2section(pe_ctx_t *ctx, uint64_t rva) {
 	return NULL;
 }
 
+// Converts a RVA (Relative Virtual Address) to a raw file offset
 uint64_t pe_rva2ofs(const pe_ctx_t *ctx, uint64_t rva) {
 	if (rva == 0 || ctx->pe.sections == NULL)
 		return 0;
 
+	// Find out which section the given RVA belongs
 	for (uint32_t i=0; i < ctx->pe.num_sections; i++) {
-		if (rva >= ctx->pe.sections[i]->VirtualAddress &&
-			rva < (ctx->pe.sections[i]->VirtualAddress
-				+ ctx->pe.sections[i]->SizeOfRawData))
-			return rva - ctx->pe.sections[i]->VirtualAddress
-				+ ctx->pe.sections[i]->PointerToRawData;
+		if (ctx->pe.sections[i] == NULL)
+			return 0;
+
+		// Use SizeOfRawData if VirtualSize == 0
+		size_t section_size = ctx->pe.sections[i]->Misc.VirtualSize;
+		if (section_size == 0)
+			section_size = ctx->pe.sections[i]->SizeOfRawData;
+
+		if (ctx->pe.sections[i]->VirtualAddress <= rva) {
+			if ((ctx->pe.sections[i]->VirtualAddress + section_size) > rva) {
+			 	rva -= ctx->pe.sections[i]->VirtualAddress;
+				rva += ctx->pe.sections[i]->PointerToRawData;
+				return rva;
+			}
+		}
 	}
-	return 0;
+
+	// Handle PE with a single section
+	if (ctx->pe.num_sections == 1) {
+		rva -= ctx->pe.sections[0]->VirtualAddress;
+		rva += ctx->pe.sections[0]->PointerToRawData;
+		return rva;
+	}
+
+	return rva; // PE with no sections, return RVA
 }
 
 // Returns the RVA for a given offset
@@ -319,12 +347,17 @@ uint64_t pe_ofs2rva(const pe_ctx_t *ctx, uint64_t ofs) {
 		return 0;
 
 	for (uint32_t i=0; i < ctx->pe.num_sections; i++) {
-		// If offset points within this section, return its VA
-		if (ofs >= ctx->pe.sections[i]->PointerToRawData &&
-			ofs < (ctx->pe.sections[i]->PointerToRawData
-				+ ctx->pe.sections[i]->SizeOfRawData))
-			return ctx->pe.sections[i]->VirtualAddress > 0 ? ofs +
-				ctx->pe.sections[i]->VirtualAddress : ofs + ctx->pe.imagebase;
+		if (ctx->pe.sections[i] == NULL)
+			return 0;
+
+		if (ctx->pe.sections[i]->PointerToRawData <= ofs) {
+			if ((ctx->pe.sections[i]->PointerToRawData + 
+			 ctx->pe.sections[i]->SizeOfRawData) > ofs) {
+			 	ofs -= ctx->pe.sections[i]->PointerToRawData;
+				ofs += ctx->pe.sections[i]->VirtualAddress;
+				return ofs;
+			}
+		}
 	}
 	return 0;
 }
@@ -368,9 +401,8 @@ IMAGE_SECTION_HEADER *pe_section_by_name(pe_ctx_t *ctx, const char *name) {
 	if (ctx->pe.sections == NULL || name == NULL)
 		return NULL;
 
-	size_t name_len = strlen(name);
 	for (uint32_t i=0; i < ctx->pe.num_sections; i++) {
-		if (memcmp(ctx->pe.sections[i]->Name, name, name_len) == 0)
+		if (strcmp((const char *)ctx->pe.sections[i]->Name, name) == 0)
 			return ctx->pe.sections[i];
 	}
 	return NULL;
