@@ -3,11 +3,11 @@
 
 	peres.c - retrive informations and binary data of resources
 
-	Copyright (C) 2012 pev authors
+	Copyright (C) 2012 - 2017 pev authors
 
 	This program is free software: you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
-	the Free Software Foundation, either version 3 of the License, or
+	the Free Software Foundation, either version 2 of the License, or
 	(at your option) any later version.
 
 	This program is distributed in the hope that it will be useful,
@@ -17,6 +17,20 @@
 
 	You should have received a copy of the GNU General Public License
 	along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+    In addition, as a special exception, the copyright holders give
+    permission to link the code of portions of this program with the
+    OpenSSL library under certain conditions as described in each
+    individual source file, and distribute linked combinations
+    including the two.
+    
+    You must obey the GNU General Public License in all respects
+    for all of the code used other than OpenSSL.  If you modify
+    file(s) with this exception, you may extend this exception to your
+    version of the file(s), but you are not obligated to do so.  If you
+    do not wish to do so, delete this exception statement from your
+    version.  If you delete this exception statement from all source
+    files in the program, then also delete it here.
 */
 
 #include "common.h"
@@ -96,17 +110,21 @@ typedef struct {
 
 static void usage(void)
 {
+	static char formats[255];
+	output_available_formats(formats, sizeof(formats), '|');
 	printf("Usage: %s OPTIONS FILE\n"
 		"Show information about resource section and extract it\n"
 		"\nExample: %s -a putty.exe\n"
 		"\nOptions:\n"
 		" -a, --all                              Show all information, statistics and extract resources\n"
+		" -f, --format <%s>  change output format (default: text)\n"
+		" -i, --info                             Show resources information\n"
+		" -s, --statistics                       Show resources statistics\n"
 		" -x, --extract                          Extract resources\n"
-		" -i, --info                             Show informations\n"
-		" -s, --statistics                       Show statistics\n"
-		" -v, --version                          Show version and exit\n"
+		" -v, --file-version                     Show File Version from PE resource directory\n"
+		" -V, --version                          show version and exit\n"
 		" --help                                 Show this help and exit\n",
-		PROGRAM, PROGRAM);
+		PROGRAM, PROGRAM, formats);
 }
 
 static void free_options(options_t *options)
@@ -123,19 +141,19 @@ static options_t *parse_options(int argc, char *argv[])
 	memset(options, 0, sizeof(options_t));
 
 	/* Parameters for getopt_long() function */
-	static const char short_options[] = "a:x:i:s:v";
+	static const char short_options[] = "a:f:isxvV";
 
 	static const struct option long_options[] = {
-		{ "all",		required_argument,	NULL, 'a' },
-		{ "extract",	no_argument,		NULL, 'x' },
-		{ "info",		no_argument,		NULL, 'i' },
-		{ "statistics",	no_argument,		NULL, 's' },
-		{ "version",	no_argument,		NULL, 'v' },
-		{ "help",		no_argument,		NULL,  1  },
-		{ NULL,			0,					NULL,  0  }
-	};
-
-	//memset(&config, false, sizeof(config));
+		{ "all",            required_argument,  NULL, 'a' },
+		{ "format",         required_argument,  NULL, 'f' },
+		{ "info",           no_argument,        NULL, 'i' },
+		{ "statistics",	    no_argument,        NULL, 's' },
+		{ "extract",	    no_argument,        NULL, 'x' },
+		{ "file-version",   no_argument,        NULL, 'v' },
+		{ "version",	    no_argument,        NULL, 'V' },
+		{ "help",           no_argument,        NULL,  1  },
+		{ NULL,             0,                  NULL,  0  }
+		};
 
 	int c, ind;
 
@@ -149,8 +167,9 @@ static options_t *parse_options(int argc, char *argv[])
 			case 'a':
 				options->all = true;
 				break;
-			case 'x':
-				options->extract = true;
+			case 'f':
+				if (output_set_format_by_name(optarg) < 0)
+					EXIT_ERROR("invalid format option");
 				break;
 			case 'i':
 				options->info = true;
@@ -158,7 +177,13 @@ static options_t *parse_options(int argc, char *argv[])
 			case 's':
 				options->statistics = true;
 				break;
+			case 'x':
+				options->extract = true;
+				break;
 			case 'v':
+				options->version = true;
+				break;
+			case 'V':
 				printf("%s %s\n%s\n", PROGRAM, TOOLKIT, COPY);
 				exit(EXIT_SUCCESS);
 			case 1: // --help option
@@ -366,7 +391,7 @@ static void saveResource(pe_ctx_t *ctx, const NODE_PERES *node)
 	const uint64_t offsetData = pe_rva2ofs(ctx, dataEntryNode->resource.dataEntry->offsetToData);
 	const size_t dataEntrySize = dataEntryNode->resource.dataEntry->size;
 	const char *buffer = LIBPE_PTR_ADD(ctx->map_addr, offsetData);
-	if (LIBPE_IS_PAST_THE_END(ctx, buffer, dataEntrySize)) {
+	if (!pe_can_read(ctx, buffer, dataEntrySize)) {
 		// TODO: Should we report something?
 		return;
 	}
@@ -421,8 +446,6 @@ static void extractResources(pe_ctx_t *ctx, const NODE_PERES *node)
 		node = node->lastNode;
 	}
 
-	output("!SAVE RESOURCES!", NULL);
-
 	while (node != NULL) {
 		if (node->nodeType != RDT_DATA_ENTRY) {
 			node = node->nextNode;
@@ -434,6 +457,64 @@ static void extractResources(pe_ctx_t *ctx, const NODE_PERES *node)
 	}
 }
 
+static void showVersion(pe_ctx_t *ctx, const NODE_PERES *node)
+{
+	assert(node != NULL);
+
+	int count = 0;
+	const NODE_PERES *dataEntryNode;
+	uint32_t nameOffset;
+	bool found = false;
+
+	while (node->lastNode != NULL) {
+		node = node->lastNode;
+	}
+
+	while (node != NULL) {
+		if (node->nodeType != RDT_DATA_ENTRY) {
+			node = node->nextNode;
+			continue;
+		}
+		count++;
+		//if (count==19)
+		dataEntryNode = lastNodeByType(node, RDT_DATA_ENTRY);
+		if (dataEntryNode == NULL)
+			return;
+		nameOffset = node->rootNode->resource.directoryEntry->DirectoryName.name.NameOffset;
+		if (nameOffset == 16) {
+			found = true;
+			break;
+		}
+		node = node->nextNode;
+	}
+
+	if (!found)
+		return;
+	
+	const uint64_t offsetData = pe_rva2ofs(ctx, dataEntryNode->resource.dataEntry->offsetToData);
+	const size_t dataEntrySize = dataEntryNode->resource.dataEntry->size;
+	const char *buffer = LIBPE_PTR_ADD(ctx->map_addr, 32 + offsetData);
+	if (!pe_can_read(ctx, buffer, dataEntrySize)) {
+		// TODO: Should we report something?
+		return;
+	}
+
+	VS_FIXEDFILEINFO *info = (VS_FIXEDFILEINFO *) buffer;
+	char value[MAX_MSG];
+
+	//snprintf(value, MAX_MSG, "%d", totalCount);
+	
+	
+	snprintf(value, MAX_MSG, "%u.%u.%u.%u",
+	(unsigned int)(info->dwProductVersionMS & 0xffff0000) >> 16,
+	(unsigned int)info->dwProductVersionMS & 0x0000ffff,
+	(unsigned int)(info->dwProductVersionLS & 0xffff0000) >> 16,
+	(unsigned int)info->dwProductVersionLS & 0x0000ffff);
+
+	output("File Version", value);
+
+}
+
 static void showInformations(const NODE_PERES *node)
 {
 	assert(node != NULL);
@@ -442,8 +523,6 @@ static void showInformations(const NODE_PERES *node)
 		node = node->lastNode;
 	}
 
-	output("!SHOW INFORMATIONS!", NULL);
-
 	while (node != NULL) {
 		showNode(node);
 		node = node->nextNode;
@@ -456,8 +535,6 @@ static void showStatistics(const NODE_PERES *node)
 		node = node->lastNode;
 	}
 
-	output("!SHOW STATISTICS!", NULL);
-
 	int totalCount = 0;
 	int totalResourceDirectory = 0;
 	int totalDirectoryEntry = 0;
@@ -534,7 +611,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 		return NULL;
 
 	uint64_t resourceDirOffset = pe_rva2ofs(ctx, resourceDirectory->VirtualAddress);
-	char s[MAX_MSG];
+	/*char s[MAX_MSG];
 
 	if (resourceDirectory->Size != 0) {
 		snprintf(s, MAX_MSG, "%#x (%d bytes)",
@@ -547,11 +624,11 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 		output(pe_directory_name(IMAGE_DIRECTORY_ENTRY_RESOURCE), s); // Resource table
 #endif
 		//printf("Offset by RVA: 0x%x\n\n", resourceDirOffset);
-	}
+	}*/
 
 	uintptr_t offset = resourceDirOffset;
 	void *ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-	if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY))) {
+	if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY))) {
 		// TODO: Should we report something?
 		return NULL;
 	}
@@ -569,7 +646,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 		offsetDirectory1 += (i == 1) ? 16 : 8;
 		offset = resourceDirOffset + offsetDirectory1;
 		ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-		if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))) {
+		if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))) {
 			// TODO: Should we report something?
 			goto _error;
 		}
@@ -587,7 +664,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 		{
 			offset = resourceDirOffset + lastDirectoryEntryNodeAtLevel1->resource.directoryEntry->DirectoryData.data.OffsetToDirectory;
 			ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-			if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY))) {
+			if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY))) {
 				// TODO: Should we report something?
 				goto _error;
 			}
@@ -604,7 +681,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 				offsetDirectory2 += (j == 1) ? 16 : 8;
 				offset = resourceDirOffset + lastNodeByTypeAndLevel(node, RDT_DIRECTORY_ENTRY, RDT_LEVEL1)->resource.directoryEntry->DirectoryData.data.OffsetToDirectory + offsetDirectory2;
 				ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-				if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))) {
+				if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))) {
 					// TODO: Should we report something?
 					goto _error;
 				}
@@ -617,7 +694,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 
 				offset = resourceDirOffset + node->resource.directoryEntry->DirectoryData.data.OffsetToDirectory; // posiciona em 0x72
 				ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-				if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY))) {
+				if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY))) {
 					// TODO: Should we report something?
 					goto _error;
 				}
@@ -634,7 +711,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 									lastNodeByTypeAndLevel(node, RDT_RESOURCE_DIRECTORY, RDT_LEVEL3)->resource.resourceDirectory->NumberOfIdEntries); y++)
 				{
 					ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-					if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))) {
+					if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DIRECTORY_ENTRY))) {
 						// TODO: Should we report something?
 						goto _error;
 					}
@@ -646,7 +723,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 
 					offset = resourceDirOffset + node->resource.directoryEntry->DirectoryName.name.NameOffset;
 					ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-					if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DATA_STRING))) {
+					if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DATA_STRING))) {
 						// TODO: Should we report something?
 						goto _error;
 					}
@@ -658,7 +735,7 @@ static NODE_PERES * discoveryNodesPeres(pe_ctx_t *ctx)
 
 					offset = resourceDirOffset + node->lastNode->resource.directoryEntry->DirectoryData.data.OffsetToDirectory;
 					ptr = LIBPE_PTR_ADD(ctx->map_addr, offset);
-					if (LIBPE_IS_PAST_THE_END(ctx, ptr, sizeof(IMAGE_RESOURCE_DATA_ENTRY))) {
+					if (!pe_can_read(ctx, ptr, sizeof(IMAGE_RESOURCE_DATA_ENTRY))) {
 						// TODO: Should we report something?
 						goto _error;
 					}
@@ -684,11 +761,16 @@ _error:
 
 int main(int argc, char **argv)
 {
+	pev_config_t config;
+	PEV_INITIALIZE(&config);
+
 	if (argc < 3) {
 		usage();
 		exit(EXIT_FAILURE);
 	}
 
+	output_set_cmdline(argc, argv);
+
 	options_t *options = parse_options(argc, argv); // opcoes
 
 	const char *path = argv[argc-1];
@@ -709,6 +791,8 @@ int main(int argc, char **argv)
 	if (!pe_is_pe(&ctx))
 		EXIT_ERROR("not a valid PE file");
 
+	output_open_document();
+
 	NODE_PERES *node = discoveryNodesPeres(&ctx);
 	if (node == NULL) {
 		fprintf(stderr, "this file has no resources\n");
@@ -719,6 +803,7 @@ int main(int argc, char **argv)
 		showInformations(node);
 		showStatistics(node);
 		extractResources(&ctx, node);
+		showVersion(&ctx, node);
 	} else {
 		if (options->extract)
 			extractResources(&ctx, node);
@@ -726,8 +811,12 @@ int main(int argc, char **argv)
 			showInformations(node);
 		if (options->statistics)
 			showStatistics(node);
+		if (options->version)
+			showVersion(&ctx, node);
 	}
 
+	output_close_document();
+
 	freeNodes(node);
 
 	// libera a memoria
@@ -740,5 +829,7 @@ int main(int argc, char **argv)
 		return EXIT_FAILURE;
 	}
 
+	PEV_FINALIZE(&config);
+
 	return EXIT_SUCCESS;
 }
