@@ -3,11 +3,11 @@
 
 	readpe.c - show PE file headers
 
-	Copyright (C) 2013 pev authors
+	Copyright (C) 2013 - 2017 pev authors
 
 	This program is free software: you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
-	the Free Software Foundation, either version 3 of the License, or
+	the Free Software Foundation, either version 2 of the License, or
 	(at your option) any later version.
 
 	This program is distributed in the hope that it will be useful,
@@ -17,17 +17,30 @@
 
 	You should have received a copy of the GNU General Public License
 	along with this program.  If not, see <http://www.gnu.org/licenses/>.
+
+    In addition, as a special exception, the copyright holders give
+    permission to link the code of portions of this program with the
+    OpenSSL library under certain conditions as described in each
+    individual source file, and distribute linked combinations
+    including the two.
+    
+    You must obey the GNU General Public License in all respects
+    for all of the code used other than OpenSSL.  If you modify
+    file(s) with this exception, you may extend this exception to your
+    version of the file(s), but you are not obligated to do so.  If you
+    do not wish to do so, delete this exception statement from your
+    version.  If you delete this exception statement from all source
+    files in the program, then also delete it here.
 */
 
 #include "common.h"
 #include <time.h>
 #include <ctype.h>
+#include "plugins.h"
 
 #define PROGRAM "readpe"
-#define MAX_DLL_NAME 256
-#define MAX_FUNCTION_NAME 512
 
-struct options {
+typedef struct {
 	bool all;
 	bool dos;
 	bool coff;
@@ -37,13 +50,12 @@ struct options {
 	bool exports;
 	bool all_headers;
 	bool all_sections;
-};
-
-static struct options config;
-static int ind;
+} options_t;
 
 static void usage(void)
 {
+	static char formats[255];
+	output_available_formats(formats, sizeof(formats), '|');
 	printf("Usage: %s OPTIONS FILE\n"
 		"Show PE file headers\n"
 		"\nExample: %s --header optional winzip.exe\n"
@@ -51,32 +63,43 @@ static void usage(void)
 		" -A, --all                              full output (default)\n"
 		" -H, --all-headers                      print all PE headers\n"
 		" -S, --all-sections                     print all PE sections headers\n"
-		" -f, --format <text|csv|xml|html>       change output format (default: text)\n"
+		" -f, --format <%s>  change output format (default: text)\n"
 		" -d, --dirs                             show data directories\n"
 		" -h, --header <dos|coff|optional>       show specific header\n"
 		" -i, --imports                          show imported functions\n"
 		" -e, --exports                          show exported functions\n"
-		" -v, --version                          show version and exit\n"
+		" -V, --version                          show version and exit\n"
 		" --help                                 show this help and exit\n",
-		PROGRAM, PROGRAM);
+		PROGRAM, PROGRAM, formats);
 }
 
-static void parse_headers(const char *optarg)
+static void parse_headers(options_t *options, const char *optarg)
 {
 	if (!strcmp(optarg, "dos"))
-		config.dos = true;
+		options->dos = true;
 	else if (!strcmp(optarg, "coff"))
-		config.coff = true;
+		options->coff = true;
 	else if (!strcmp(optarg, "optional"))
-		config.opt = true;
+		options->opt = true;
 	else
 		EXIT_ERROR("invalid header option");
 }
 
-void parse_options(int argc, char *argv[])
+static void free_options(options_t *options)
 {
+	if (options == NULL)
+		return;
+
+	free(options);
+}
+
+static options_t *parse_options(int argc, char *argv[])
+{
+	options_t *options = malloc_s(sizeof(options_t));
+	memset(options, 0, sizeof(options_t));
+
 	/* Parameters for getopt_long() function */
-	static const char short_options[] = "AHSh:dief:v";
+	static const char short_options[] = "AHSh:dief:V";
 
 	static const struct option long_options[] = {
 		{ "help",             no_argument,       NULL,  1  },
@@ -88,16 +111,14 @@ void parse_options(int argc, char *argv[])
 		{ "exports",          no_argument,       NULL, 'e' },
 		{ "dirs",             no_argument,       NULL, 'd' },
 		{ "format",           required_argument, NULL, 'f' },
-		{ "version",          no_argument,       NULL, 'v' },
+		{ "version",          no_argument,       NULL, 'V' },
 		{  NULL,              0,                 NULL,  0  }
 	};
 
-	// setting all fields to false
-	memset(&config, false, sizeof(config));
+	options->all = true;
 
-	config.all = true;
+	int c, ind;
 
-	int c;
 	while ((c = getopt_long(argc, argv, short_options, long_options, &ind)))
 	{
 		if (c < 0)
@@ -109,43 +130,46 @@ void parse_options(int argc, char *argv[])
 				usage();
 				exit(EXIT_SUCCESS);
 			case 'A':
-				config.all = true;
+				options->all = true;
 				break;
 			case 'H':
-				config.all = false;
-				config.all_headers = true;
+				options->all = false;
+				options->all_headers = true;
 				break;
 			case 'd':
-				config.all = false;
-				config.dirs = true;
+				options->all = false;
+				options->dirs = true;
 				break;
 			case 'S':
-				config.all = false;
-				config.all_sections = true;
+				options->all = false;
+				options->all_sections = true;
 				break;
-			case 'v':
+			case 'V':
 				printf("%s %s\n%s\n", PROGRAM, TOOLKIT, COPY);
 				exit(EXIT_SUCCESS);
 			case 'h':
-				config.all = false;
-				parse_headers(optarg);
+				options->all = false;
+				parse_headers(options, optarg);
 				break;
 			case 'i':
-				config.all = false;
-				config.imports = true;
+				options->all = false;
+				options->imports = true;
 				break;
 			case 'e':
-				config.all = false;
-				config.exports = true;
+				options->all = false;
+				options->exports = true;
 				break;
 			case 'f':
-				parse_format(optarg);
+				if (output_set_format_by_name(optarg) < 0)
+					EXIT_ERROR("invalid format option");
 				break;
 			default:
 				fprintf(stderr, "%s: try '--help' for more information\n", PROGRAM);
 				exit(EXIT_FAILURE);
 		}
 	}
+
+	return options;
 }
 
 static void print_sections(pe_ctx_t *ctx)
@@ -183,7 +207,7 @@ static void print_sections(pe_ctx_t *ctx)
 	};
 	static const size_t max_flags = LIBPE_SIZEOF_ARRAY(valid_flags);
 
-	output("Sections", NULL);
+	output_open_scope("Sections", OUTPUT_SCOPE_TYPE_ARRAY);
 
 	const uint32_t num_sections = pe_sections_count(ctx);
 	if (num_sections == 0 || num_sections > MAX_SECTIONS)
@@ -197,6 +221,8 @@ static void print_sections(pe_ctx_t *ctx)
 
 	for (uint32_t i=0; i < num_sections; i++)
 	{
+		output_open_scope("Section", OUTPUT_SCOPE_TYPE_OBJECT);
+
 		snprintf(s, MAX_MSG, "%s", sections[i]->Name);
 		output("Name", s);
 
@@ -219,6 +245,8 @@ static void print_sections(pe_ctx_t *ctx)
 		snprintf(s, MAX_MSG, "%#x", sections[i]->Characteristics);
 		output("Characteristics", s);
 
+		output_open_scope("Characteristic Names", OUTPUT_SCOPE_TYPE_ARRAY);
+
 		for (size_t j=0; j < max_flags; j++) {
 			if (sections[i]->Characteristics & valid_flags[j]) {
 #ifdef LIBPE_ENABLE_OUTPUT_COMPAT_WITH_V06
@@ -230,7 +258,12 @@ static void print_sections(pe_ctx_t *ctx)
 			}
 		}
 
+		output_close_scope(); // Characteristic Names
+
+		output_close_scope(); // Section
 	}
+
+	output_close_scope(); // Sections
 }
 
 static void print_directories(pe_ctx_t *ctx)
@@ -260,7 +293,7 @@ static void print_directories(pe_ctx_t *ctx)
 	};
 	//static const size_t max_directory_entry = LIBPE_SIZEOF_ARRAY(names);
 #endif
-	output("Data directories", NULL);
+	output_open_scope("Data directories", OUTPUT_SCOPE_TYPE_ARRAY);
 
 	const uint32_t num_directories = pe_directories_count(ctx);
 	if (num_directories == 0 || num_directories > MAX_DIRECTORIES)
@@ -274,6 +307,7 @@ static void print_directories(pe_ctx_t *ctx)
 
 	for (uint32_t i=0; i < num_directories; i++) {
 		if (directories[i]->Size) {
+			output_open_scope("Directory", OUTPUT_SCOPE_TYPE_OBJECT);
 			snprintf(s, MAX_MSG, "%#x (%d bytes)",
 					directories[i]->VirtualAddress,
 					directories[i]->Size);
@@ -282,8 +316,11 @@ static void print_directories(pe_ctx_t *ctx)
 #else
 			output(pe_directory_name(i), s);
 #endif
+			output_close_scope(); // Directory
 		}
 	}
+
+	output_close_scope(); // Data directories
 }
 
 static void print_optional_header(IMAGE_OPTIONAL_HEADER *header)
@@ -316,10 +353,10 @@ static void print_optional_header(IMAGE_OPTIONAL_HEADER *header)
 	if (!header)
 		return;
 
-	output("Optional/Image header", NULL);
-
 	char s[MAX_MSG];
 
+	output_open_scope("Optional/Image header", OUTPUT_SCOPE_TYPE_OBJECT);
+
 	switch (header->type)
 	{
 		case MAGIC_PE32:
@@ -406,10 +443,14 @@ static void print_optional_header(IMAGE_OPTIONAL_HEADER *header)
 			output("DLL characteristics", s);
 
 #ifndef LIBPE_ENABLE_OUTPUT_COMPAT_WITH_V06
+			output_open_scope("DLL characteristics names", OUTPUT_SCOPE_TYPE_ARRAY);
+
 			for (uint16_t i=0, flag=0x0001; i < 16; i++, flag <<= 1) {
 				if (header->_32->DllCharacteristics & flag)
 					output(NULL, pe_image_dllcharacteristic_name(flag));
 			}
+
+			output_close_scope(); // DLL characteristics names
 #endif
 
 			snprintf(s, MAX_MSG, "%#x", header->_32->SizeOfStackReserve);
@@ -506,10 +547,14 @@ static void print_optional_header(IMAGE_OPTIONAL_HEADER *header)
 			output("DLL characteristics", s);
 
 #ifndef LIBPE_ENABLE_OUTPUT_COMPAT_WITH_V06
+			output_open_scope("DLL characteristics names", OUTPUT_SCOPE_TYPE_ARRAY);
+
 			for (uint16_t i=0, flag=0x0001; i < 16; i++, flag <<= 1) {
 				if (header->_64->DllCharacteristics & flag)
 					output(NULL, pe_image_dllcharacteristic_name(flag));
 			}
+
+			output_close_scope(); // DLL characteristics names
 #endif
 
 			snprintf(s, MAX_MSG, "%#"PRIx64, header->_64->SizeOfStackReserve);
@@ -526,6 +571,8 @@ static void print_optional_header(IMAGE_OPTIONAL_HEADER *header)
 			break;
 		}
 	}
+
+	output_close_scope(); // Optional/Image heade
 }
 
 static void print_coff_header(IMAGE_COFF_HEADER *header)
@@ -585,7 +632,7 @@ static void print_coff_header(IMAGE_COFF_HEADER *header)
 	static const size_t max_machine_type = LIBPE_SIZEOF_ARRAY(machineTypeTable);
 #endif
 
-	output("COFF/File header", NULL);
+	output_open_scope("COFF/File header", OUTPUT_SCOPE_TYPE_OBJECT);
 
 #ifdef LIBPE_ENABLE_OUTPUT_COMPAT_WITH_V06
 	const char *machine = "Unknown machine type";
@@ -625,6 +672,8 @@ static void print_coff_header(IMAGE_COFF_HEADER *header)
 	snprintf(s, MAX_MSG, "%#x", header->Characteristics);
 	output("Characteristics", s);
 
+	output_open_scope("Characteristics names", OUTPUT_SCOPE_TYPE_ARRAY);
+
 	for (uint16_t i=0, flag=0x0001; i < 16; i++, flag <<= 1) {
 		if (header->Characteristics & flag)
 #ifdef LIBPE_ENABLE_OUTPUT_COMPAT_WITH_V06
@@ -633,13 +682,17 @@ static void print_coff_header(IMAGE_COFF_HEADER *header)
 			output(NULL, pe_image_characteristic_name(flag));
 #endif
 	}
+
+	output_close_scope(); // Characteristics names
+
+	output_close_scope(); // COFF/File header
 }
 
 static void print_dos_header(IMAGE_DOS_HEADER *header)
 {
 	char s[MAX_MSG];
 
-	output("DOS Header", NULL);
+	output_open_scope("DOS Header", OUTPUT_SCOPE_TYPE_OBJECT);
 
 	snprintf(s, MAX_MSG, "%#x (MZ)", header->e_magic);
 	output("Magic number", s);
@@ -688,6 +741,8 @@ static void print_dos_header(IMAGE_DOS_HEADER *header)
 
 	snprintf(s, MAX_MSG, "%#x", header->e_lfanew);
 	output("PE header offset", s);
+
+	output_close_scope(); // DOS Header
 }
 
 static void print_imported_functions(pe_ctx_t *ctx, uint64_t offset)
@@ -706,7 +761,7 @@ static void print_imported_functions(pe_ctx_t *ctx, uint64_t offset)
 			case MAGIC_PE32:
 			{
 				const IMAGE_THUNK_DATA32 *thunk = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-				if (LIBPE_IS_PAST_THE_END(ctx, thunk, sizeof(IMAGE_THUNK_DATA32))) {
+				if (!pe_can_read(ctx, thunk, sizeof(IMAGE_THUNK_DATA32))) {
 					// TODO: Should we report something?
 					return;
 				}
@@ -724,13 +779,15 @@ static void print_imported_functions(pe_ctx_t *ctx, uint64_t offset)
 				} else {
 					const uint64_t imp_ofs = pe_rva2ofs(ctx, thunk->u1.AddressOfData);
 					const IMAGE_IMPORT_BY_NAME *imp_name = LIBPE_PTR_ADD(ctx->map_addr, imp_ofs);
-					if (LIBPE_IS_PAST_THE_END(ctx, imp_name, sizeof(IMAGE_IMPORT_BY_NAME))) {
+					if (!pe_can_read(ctx, imp_name, sizeof(IMAGE_IMPORT_BY_NAME))) {
 						// TODO: Should we report something?
 						return;
 					}
 
 					snprintf(hint_str, sizeof(hint_str)-1, "%d", imp_name->Hint);
 					strncpy(fname, (char *)imp_name->Name, sizeof(fname)-1);
+					// Because `strncpy` does not guarantee to NUL terminate the string itself, this must be done explicitly.
+					fname[sizeof(fname) - 1] = '\0';
 					//size_t fname_len = strlen(fname);
 				}
 				ofs += sizeof(IMAGE_THUNK_DATA32);
@@ -739,7 +796,7 @@ static void print_imported_functions(pe_ctx_t *ctx, uint64_t offset)
 			case MAGIC_PE64:
 			{
 				const IMAGE_THUNK_DATA64 *thunk = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-				if (LIBPE_IS_PAST_THE_END(ctx, thunk, sizeof(IMAGE_THUNK_DATA64))) {
+				if (!pe_can_read(ctx, thunk, sizeof(IMAGE_THUNK_DATA64))) {
 					// TODO: Should we report something?
 					return;
 				}
@@ -757,13 +814,15 @@ static void print_imported_functions(pe_ctx_t *ctx, uint64_t offset)
 				} else {
 					uint64_t imp_ofs = pe_rva2ofs(ctx, thunk->u1.AddressOfData);
 					const IMAGE_IMPORT_BY_NAME *imp_name = LIBPE_PTR_ADD(ctx->map_addr, imp_ofs);
-					if (LIBPE_IS_PAST_THE_END(ctx, imp_name, sizeof(IMAGE_IMPORT_BY_NAME))) {
+					if (!pe_can_read(ctx, imp_name, sizeof(IMAGE_IMPORT_BY_NAME))) {
 						// TODO: Should we report something?
 						return;
 					}
 
 					snprintf(hint_str, sizeof(hint_str)-1, "%d", imp_name->Hint);
 					strncpy(fname, (char *)imp_name->Name, sizeof(fname)-1);
+					// Because `strncpy` does not guarantee to NUL terminate the string itself, this must be done explicitly.
+					fname[sizeof(fname) - 1] = '\0';
 					//size_t fname_len = strlen(fname);
 				}
 				ofs += sizeof(IMAGE_THUNK_DATA64);
@@ -771,7 +830,14 @@ static void print_imported_functions(pe_ctx_t *ctx, uint64_t offset)
 			}
 		}
 
-		output(NULL, is_ordinal ? hint_str : fname);
+		output_open_scope("Function", OUTPUT_SCOPE_TYPE_OBJECT);
+
+		if (is_ordinal)
+			output("Ordinal", hint_str);
+		else
+			output("Name", fname);
+
+		output_close_scope(); // Function
 	}
 }
 
@@ -791,14 +857,14 @@ static void print_exports(pe_ctx_t *ctx)
 
 	ofs = pe_rva2ofs(ctx, va);
 	const IMAGE_EXPORT_DIRECTORY *exp = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-	if (LIBPE_IS_PAST_THE_END(ctx, exp, sizeof(IMAGE_EXPORT_DIRECTORY))) {
+	if (!pe_can_read(ctx, exp, sizeof(IMAGE_EXPORT_DIRECTORY))) {
 		// TODO: Should we report something?
 		return;
 	}
 
 	ofs = pe_rva2ofs(ctx, exp->AddressOfNames);
 	const uint32_t *rva_ptr = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-	if (LIBPE_IS_PAST_THE_END(ctx, rva_ptr, sizeof(uint32_t))) {
+	if (!pe_can_read(ctx, rva_ptr, sizeof(uint32_t))) {
 		// TODO: Should we report something?
 		return;
 	}
@@ -806,30 +872,114 @@ static void print_exports(pe_ctx_t *ctx)
 
 	ofs = pe_rva2ofs(ctx, rva);
 
-	output("Exported functions", NULL);
-	for (uint32_t i=0; i < exp->NumberOfNames; i++) {
-		const uint64_t aux = ofs; // Store current ofs
+	output_open_scope("Exported functions", OUTPUT_SCOPE_TYPE_ARRAY);
+	// If `NumberOfNames == 0` then all functions are exported by ordinal.
+	// Otherwise `NumberOfNames` must be equal to `NumberOfFunctions`
+	if (exp->NumberOfNames != 0 && exp->NumberOfNames != exp->NumberOfFunctions) {
+		fprintf(stderr, "NumberOfFunctions differs from NumberOfNames\n");
+		output_close_scope(); // Exported functions
+	}
 
-		ofs = exp->AddressOfFunctions + sizeof(uint32_t) * i;
-		uint32_t *faddr_ptr = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-		if (LIBPE_IS_PAST_THE_END(ctx, faddr_ptr, sizeof(uint32_t))) {
+	uint64_t offset_to_AddressOfFunctions = pe_rva2ofs(ctx, exp->AddressOfFunctions);
+	uint64_t offset_to_AddressOfNames = pe_rva2ofs(ctx, exp->AddressOfNames);
+	uint64_t offset_to_AddressOfNameOrdinals = pe_rva2ofs(ctx, exp->AddressOfNameOrdinals);
+
+	//
+	// The format of IMAGE_EXPORT_DIRECTORY can be seen in http://i.msdn.microsoft.com/dynimg/IC60608.gif
+	//
+
+	// We want to use `NumberOfFunctions` for looping as it's the total number of functions/symbols
+	// exported by the module. On the other hand, `NumberOfNames` is the number of
+	// functions/symbols exported by name only.
+	for (uint32_t i=0; i < exp->NumberOfFunctions; i++) {
+		uint64_t entry_ordinal_list_ptr = offset_to_AddressOfNameOrdinals + sizeof(uint16_t) * i;
+		uint16_t *entry_ordinal_list = LIBPE_PTR_ADD(ctx->map_addr, entry_ordinal_list_ptr);
+
+		uint64_t entry_va_list_ptr = offset_to_AddressOfFunctions + sizeof(uint32_t) * i;
+		uint32_t *entry_va_list = LIBPE_PTR_ADD(ctx->map_addr, entry_va_list_ptr);
+
+		uint64_t entry_name_list_ptr = offset_to_AddressOfNames + sizeof(uint32_t) * i;
+		uint32_t *entry_name_list = LIBPE_PTR_ADD(ctx->map_addr, entry_name_list_ptr);
+
+		// printf("ctx->map_addr = %p\n", ctx->map_addr);
+		// printf("ctx->map_end = %p\n", ctx->map_end);
+		// printf("entry_ordinal_list = %p\n", entry_ordinal_list);
+		// printf("entry_va_list = %p\n", entry_va_list);
+		// printf("entry_name_list = %p\n", entry_name_list);
+
+		if (!pe_can_read(ctx, entry_ordinal_list, sizeof(uint32_t))) {
 			// TODO: Should we report something?
 			break;
 		}
-		const uint32_t faddr = *faddr_ptr;
 
-		ofs = aux; // Restore previous ofs
+		if (!pe_can_read(ctx, entry_va_list, sizeof(uint32_t))) {
+			// TODO: Should we report something?
+			break;
+		}
 
-		char addr[30];
-		snprintf(addr, 30, "%#x", faddr);
+		if (!pe_can_read(ctx, entry_name_list, sizeof(uint32_t))) {
+			// TODO: Should we report something?
+			break;
+		}
 
-		const char *fname_ptr = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-		// TODO: Validate if it's ok to read fname_ptr+N
-		char fname[300];
-		strncpy(fname, fname_ptr, sizeof(fname)-1);
+		// Add `Base` to the element of `AddressOfNameOrdinals` array to get the correct ordinal..
+		//const uint16_t entry_ordinal = exp->Base + *entry_ordinal_list;
+		const uint32_t entry_va = *entry_va_list;
+		const uint32_t entry_name_rva = *entry_name_list;
+		const uint64_t entry_name_ofs = pe_rva2ofs(ctx, entry_name_rva);
+		const char *entry_name = LIBPE_PTR_ADD(ctx->map_addr, entry_name_ofs);
 
-		output(addr, fname);
+		// Validate whether it's ok to access at least 1 byte after entry_name.
+		// It might be '\0', for example.
+		if (!pe_can_read(ctx, entry_name, 1)) {
+			// TODO: Should we report something?
+			break;
+		}
+
+		//printf("ord=%d, va=%x, name=%s\n", entry_ordinal, entry_va, entry_name);
+
+		// Declared as 11 bytes so that it can store the hexadecimal representation of the maximum
+		// possible value of an uint32_t variable, 0xFFFFFFFF.
+		char addr[11] = { 0 };
+		sprintf(addr, "%#x", entry_va);
+
+		char fname[300] = { 0 };
+		strncpy(fname, entry_name, sizeof(fname)-1);
+		// Because `strncpy` does not guarantee to NUL terminate the string itself, this must be done explicitly.
+		fname[sizeof(fname) - 1] = '\0';
+
+		output_open_scope("Function", OUTPUT_SCOPE_TYPE_OBJECT);
+
+		// Check whether the exported function is forwarded.
+		// It's forwarded if its RVA is inside the exports section.
+		if (entry_va >= va && entry_va <= va + dir->Size)
+		{
+			// When a symbol is forwarded, its RVA points to a string containing
+			// the name of the DLL and symbol to which it is forwarded.
+			const uint64_t fw_entry_name_ofs = pe_rva2ofs(ctx, entry_va);
+			const char *fw_entry_name = LIBPE_PTR_ADD(ctx->map_addr, fw_entry_name_ofs);
+
+			// Validate whether it's ok to access at least 1 byte after fw_entry_name.
+			// It might be '\0', for example.
+			if (!pe_can_read(ctx, fw_entry_name, 1)) {
+				// TODO: Should we report something?
+				break;
+			}
+
+			char fname_forwarded[sizeof(fname) * 2 + 4] = { 0 }; // Twice the size plus " -> ".
+			snprintf(fname_forwarded, sizeof(fname_forwarded)-1, "%s -> %s", fname, fw_entry_name);
+
+			output(addr, fname_forwarded);
+		}
+		else
+		{
+			output(addr, fname);
+		}
+
+		output_close_scope(); // Function
 	}
+
+	output_close_scope(); // Exported functions
 }
 
 static void print_imports(pe_ctx_t *ctx)
@@ -843,15 +993,15 @@ static void print_imports(pe_ctx_t *ctx)
 		fprintf(stderr, "import directory not found\n");
 		return;
 	}
-
 	uint64_t ofs = pe_rva2ofs(ctx, va);
 
-	output("Imported functions", NULL);
+	output_open_scope("Imported functions", OUTPUT_SCOPE_TYPE_ARRAY);
 
 	while (1) {
 		IMAGE_IMPORT_DESCRIPTOR *id = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-		if (LIBPE_IS_PAST_THE_END(ctx, id, sizeof(IMAGE_IMPORT_DESCRIPTOR))) {
+		if (!pe_can_read(ctx, id, sizeof(IMAGE_IMPORT_DESCRIPTOR))) {
 			// TODO: Should we report something?
+			output_close_scope();
 			return;
 		}
 
@@ -864,32 +1014,57 @@ static void print_imports(pe_ctx_t *ctx)
 		ofs = pe_rva2ofs(ctx, id->Name);
 		if (ofs == 0)
 			break;
+
 		const char *dll_name_ptr = LIBPE_PTR_ADD(ctx->map_addr, ofs);
-		// TODO: Validate if it's ok to read dll_name_ptr+N
+		// Validate whether it's ok to access at least 1 byte after dll_name_ptr.
+		// It might be '\0', for example.
+		if (!pe_can_read(ctx, dll_name_ptr, 1)) {
+			// TODO: Should we report something?
+			break;
+		}
+
 		char dll_name[MAX_DLL_NAME];
 		strncpy(dll_name, dll_name_ptr, sizeof(dll_name)-1);
+		// Because `strncpy` does not guarantee to NUL terminate the string itself, this must be done explicitly.
+		dll_name[sizeof(dll_name) - 1] = '\0';
 
-		output(dll_name, NULL);
+		output_open_scope("Library", OUTPUT_SCOPE_TYPE_OBJECT);
+		output("Name", dll_name);
 
 		ofs = pe_rva2ofs(ctx, id->u1.OriginalFirstThunk ? id->u1.OriginalFirstThunk : id->FirstThunk);
-		if (ofs == 0)
+		if (ofs == 0) {
+			output_close_scope(); // Library
 			break;
+		}
+
+		output_open_scope("Functions", OUTPUT_SCOPE_TYPE_ARRAY);
 
 		// Search for DLL imported functions
 		print_imported_functions(ctx, ofs);
 
+		output_close_scope(); // Functions
+
 		ofs = aux; // Restore previous ofs
+
+		output_close_scope(); // Library
 	}
+
+	output_close_scope(); // Imported functions
 }
 
 int main(int argc, char *argv[])
 {
+	pev_config_t config;
+	PEV_INITIALIZE(&config);
+
 	if (argc < 2) {
 		usage();
 		return EXIT_FAILURE;
 	}
 
-	parse_options(argc, argv); // Opcoes
+	output_set_cmdline(argc, argv);
+
+	options_t *options = parse_options(argc, argv); // opcoes
 
 	pe_ctx_t ctx;
 
@@ -908,8 +1083,10 @@ int main(int argc, char *argv[])
 	if (!pe_is_pe(&ctx))
 		EXIT_ERROR("not a valid PE file");
 
+	output_open_document();
+
 	// dos header
-	if (config.dos || config.all_headers || config.all) {
+	if (options->dos || options->all_headers || options->all) {
 		IMAGE_DOS_HEADER *header_ptr = pe_dos(&ctx);
 		if (header_ptr)
 			print_dos_header(header_ptr);
@@ -917,7 +1094,7 @@ int main(int argc, char *argv[])
 	}
 
 	// coff/file header
-	if (config.coff || config.all_headers || config.all) {
+	if (options->coff || options->all_headers || options->all) {
 		IMAGE_COFF_HEADER *header_ptr = pe_coff(&ctx);
 		if (header_ptr)
 			print_coff_header(header_ptr);
@@ -925,41 +1102,58 @@ int main(int argc, char *argv[])
 	}
 
 	// optional header
-	if (config.opt || config.all_headers || config.all) {
+	if (options->opt || options->all_headers || options->all) {
 		IMAGE_OPTIONAL_HEADER *header_ptr = pe_optional(&ctx);
 		if (header_ptr)
 			print_optional_header(header_ptr);
 		else { EXIT_ERROR("unable to read Optional (Image) file header"); }
 	}
 
+	IMAGE_DATA_DIRECTORY **directories = pe_directories(&ctx);
+	bool directories_warned = false;
+
 	// directories
-	if (config.dirs || config.all) {
-		if (pe_directories(&ctx) != NULL)
+	if (options->dirs || options->all) {
+		if (directories != NULL)
 			print_directories(&ctx);
-		else { EXIT_ERROR("unable to read the Directories entry from Optional header"); }
+		else if (!directories_warned) {
+			fprintf(stderr, "directories not found\n");
+			directories_warned = true;
+		}
 	}
 
 	// imports
-	if (config.imports || config.all) {
-		if (pe_directories(&ctx) != NULL)
+	if (options->imports || options->all) {
+		if (directories != NULL)
 			print_imports(&ctx);
-		else { EXIT_ERROR("unable to read the Directories entry from Optional header"); }
+		else if (!directories_warned) {
+			fprintf(stderr, "directories not found\n");
+			directories_warned = true;
+		}
 	}
 
 	// exports
-	if (config.exports || config.all) {
-		if (pe_directories(&ctx) != NULL)
+	if (options->exports || options->all) {
+		if (directories != NULL)
 			print_exports(&ctx);
-		else { EXIT_ERROR("unable to read directories from optional header"); }
+		else if (!directories_warned) {
+			fprintf(stderr, "directories not found\n");
+			directories_warned = true;
+		}
 	}
 
 	// sections
-	if (config.all_sections || config.all) {
+	if (options->all_sections || options->all) {
 		if (pe_sections(&ctx) != NULL)
 			print_sections(&ctx);
 		else { EXIT_ERROR("unable to read sections"); }
 	}
 
+	output_close_document();
+
+	// libera a memoria
+	free_options(options);
+
 	// free
 	err = pe_unload(&ctx);
 	if (err != LIBPE_E_OK) {
@@ -967,5 +1161,7 @@ int main(int argc, char *argv[])
 		return EXIT_FAILURE;
 	}
 
+	PEV_FINALIZE(&config);
+
 	return EXIT_SUCCESS;
 }
